# -------------------------------------------------------------------------
# Copyright (c) 2023 PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

r"""`ua_gateway` module provides support for Kepware's UA Gateway plug-in 
specific objects within the Kepware Configuration API
"""

from . import certificates, client, server