# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

r"""`admin` module provides functionality to manage Kepware Administration based 
properties available through the Kepware Configuration API
"""

from . import users, user_groups,ua_server, lls