# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

r"""`udd` module provides support for Universal Device driver specific objects 
within the Kepware Configuration API
"""

from . import profile